/* test/example/benchmark for nfa.c
 * pesco 2021-2022, isc license
 */

#include "nfa.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>


/* the parser's mind */

enum gedanken {
	BEGWORD = INNER
};
enum ausgaben {
	WORD = OUTPUT,
	ATSIGN
};

struct mind {
	const char *input, *begword, *word, *atsign;
};

int
sub(int symbol, size_t state, int i, void *env_)
{
	struct mind *env = env_;

	assert(env->input != NULL);

	switch (symbol) {
	case BEGWORD:
		//printf("%8s[%3zu] pos %d\n", "BEGWORD", state, i);
		env->begword = env->input + i;
		return 1;
	}

	return 0;	/* my mind is blank */
}

int
out_print(struct range r, size_t state, int i, void *env_)
{
	int l;
	struct mind *env = env_;

	assert(env->input != NULL);

	/* inward output - actions */
	assert(r.n == 1);
	switch (r.base) {
	case ATSIGN:
		env->atsign = env->input + i;
		printf("%8s[%3zu] pos %d\n", "ATSIGN", state, i - 1);
		return r.base;
	case WORD:
		env->word = env->input + i;
		assert(env->begword != NULL);
		assert(env->begword < env->word);
		l = env->word - env->begword;
		printf("%8s[%3zu] pos %d\t%.*s\n", "WORD", state, i - l,
		    l, env->begword);
		return r.base;
	}

	return -1;	/* what can i do? */
}

int
out_bench(struct range r, size_t state, int i, void *env_)
{
	struct mind *env = env_;

	assert(env->input != NULL);

	/* inward output - actions */
	assert(r.n == 1);
	switch (r.base) {
	case ATSIGN:
		env->atsign = env->input + i;
		return r.base;
	case WORD:
		env->word = env->input + i;
		assert(env->begword != NULL);
		assert(env->begword < env->word);
		return r.base;
	}

	return -1;	/* what can i do? */
}


#include <time.h>	/* clock_gettime */
#include <sys/time.h>	/* timespecsub */

int
benchmark(NFA a)
{
	static const char *inputs[] = {
	    "test@example.com",
	    "hans.wurst+kaese@gmail.com",
	    "!#$%&@valid.email",
	    "...@invalid",
	    "total@falsch.",
	    "auch@.falsch",
	    "genauso.@falsch.ey",
	    "bis@hier.richtig ABER",
	    "mett@wurst",
	    "hallo@",
	    "@dingens",
	    "@",
	    "",
	    NULL };
	
	static const long N = 10000;		/* runs per input */

	struct prep *pr;
	struct mind env;
	struct timespec t, t0, td;
	const char **p;
	int i, n;
	size_t sz;

	env.input = inputs[0];
	pr = nfaprep(a, sub, out_bench, &env);

	/* warmup */
	sz = strlen(inputs[0]);
	for (i = 0; i < N; i++)
		nfarun(pr, inputs[0], sz);

	for (p = inputs; *p != NULL; p++) {
		printf("%30.30s  ", *p);
		fflush(stdout);

		sz = strlen(*p);
		env.input = *p;
		clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &t0);
		for (i = 0; i < N; i++)
			n = nfarun(pr, *p, sz);
		clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &t);
		timespecsub(&t, &t0, &td);

		if (n == sz)
			printf("match     ");
		else if (n < 0)
			printf("no match  ");
		else
			printf("partial   ");

		printf("%9.0f ns\n", (td.tv_sec * 1e9 + td.tv_nsec) / N);
	}

	return 0;
}


#include <getopt.h>

int
main(int argc, char **argv)
{
	/* character classes */
	NFA lower = range('a', 'z');
	NFA upper = range('A', 'Z');
	NFA digit = range('0', '9');
	NFA alpha = choice(lower, upper);
	NFA other = choice(chr('!'), range('#', '\''), range('*', '-'),
	    chr('/'), chr('='), chr('?'), range('^', '`'), range('{', '~'));
		/* = oneof("!#$%&'*+-/=?^_`{|}~") */

	/* email addresses. cf. rfc 5322, section 3.4 */
	NFA atext	= choice(alpha, digit, other);
	NFA word	= seq(think(BEGWORD), plus(atext), output(WORD));
	NFA atsign	= seq(chr('@'), output(ATSIGN));
	NFA dotted	= seq(word, star(seq(chr('.'), word)));
	NFA address	= seq(dotted, atsign, dotted);

	NFA nfa = address;
	int ch;

	while ((ch = getopt(argc, argv, "b")) != -1) {
		switch (ch) {
		case 'b':
			return benchmark(nfa);
		default:
			printf("usage: nfa [-b] string ...\n");
			return 1;
		}
	}

	nfaprint(nfa);
	putchar('\n');
	for (int i = 1; i < argc; i++) {
		char *inp;
		size_t sz;
		int n;
		struct mind env;

		inp = argv[i];
		sz = strlen(inp);
		env.input = inp;
		n = nfamatch(nfa, sub, out_print, &env, inp, sz);
		printf("\n");

		if (n == -1)
			printf("no ");
		else if (n < sz)
			printf("partial ");
		printf("match: %.*s\n", n, inp);
	}

	return 0;
}
